=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed April 2015 by Fredo6

# Permission to use this software for any purpose and without fee is hereby granted
# Distribution of this software for any purpose is subject to:
#  - the expressed, written consent of the author
#  - the inclusion of the present copyright notice in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   body_Lib6DeepMover.rb
# Original Date	:   27 Mar 15
# Description	:   Manage movements of grouponents at deep level (implementation)
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

module Traductor

T6[:TIP_DeepMover_Repeat] = "Repeat the last Movement"
T6[:TIP_DeepMover_CancelSelection] = "Clear the Current selection"
T6[:TIP_DeepMover_RedoSelection] = "Restore the Previous selection"
T6[:MSG_DeepMover_Selection] = "Selection of Objects - Click and drop-down"
T6[:TIP_DeepMover_DirectionTranslation] = "Direction for Translation"
T6[:TIP_DeepMover_NormalRotation] = "Axe for Rotation or Spin"
T6[:TIP_SwitchToPitch] = "Switch to Pitch mode (distance for a single rotation)"
T6[:TIP_SwitchToTPI] = "Switch to TPI mode (Threads by Inch)"
T6[:MSG_DeepMoverPivotRef] = "Rotation pivot by reference to object in Orange"
T6[:MSG_DeepMoverPivotAbs] = "Rotation pivot by reference to Model axes"
T6[:MNU_DeepMoverPivotRef] = "Pivot by Ref to:"
T6[:VCB_HintPitch] = "VCB: distance in model units followed by *"	
T6[:VCB_HintTPI] = "VCB: number of rotations followed by /"	

#=============================================================================================
#=============================================================================================
# Class DeepMover: Manage interactive selection at any hierarchical level
#=============================================================================================
#=============================================================================================

class DeepMover

#INIT: Instance initialization
def initialize__(*hargs)	
	#Parsing the arguments
	hargs.each { |arg| arg.each { |key, value|  parse_args(key, value) } if arg.class == Hash }	
	
	#Basic initialization
	@model = Sketchup.active_model
	@view = @model.active_view
	@tr_id = Geom::Transformation.new
	@ph = @view.pick_helper
	@ogl = Traductor::OpenGL_6.new

	#Create the Origin-Target Picker
	hsh = { :notify_proc => self.method("notify_from_otpicker"), :ignore_selection => true, :use_comp_for_local => true }
	@otpicker = Traductor::OriginTargetPicker.new nil, hsh
	
	#Initializing the VCB
	vcb_init
	
	#Instantiating the PseudoComp manager
	@pcompman = Traductor::PseudoCompManager.instantiate unless @pcompman

	#Creating the deep selector
	hsh = { :keyman => @keyman, :pcompman => @pcompman, :notify_proc => self.method("notify_from_selector") }
	@deep_selector = Traductor::DeepSelector.new hsh
	
	#Reset environment
	init_colors
	init_texts
	
	#Creating the protractor shape
	hsh = { :show_angle => true }
	@protractor = G6::ShapeProtractor.new hsh

	#Param bounding value
	@explode_factor_min = 1.0
	@explode_factor_max = 8.0
	
	#Initialize once
	@hsh_params = {}
	@hsh_repeat = {}
	@hsh_params_save = {}
	reset
	@cur_tool = :movement_selection
end

#INIT: Parse the arguments of the initialize method
def parse_args(key, value)
	skey = key.to_s
	case skey
	when /keyman/i
		@keyman = value
	when /pcompman/i
		@pcompman = value
	when /notify_proc/i
		@notify_proc = value
	when /param_use_screw_tpi/	
		@param_use_screw_tpi = value
	end	
end	

#INIT: set the screw TPI mode (versus Pitch mode)
def set_screw_tpi(on)
	@param_use_screw_tpi = on
	if @hsh_allowed
		@hsh_allowed[:tpi] = on
		@hsh_allowed[:pitch] = !on
	end	
end
	
#INIT: Notify the caller about events
def notify_caller(event, *args)
	puts "EVENT DEEP MOVER #{event}"
	case event
	when :selection_changed
		@deep_selector.reset_drawing
		unless @edit_mode
			@no_redo = false
			@notify_proc.call(event, *args) if @notify_proc
			transformation_apply_at_selection
		end
		
	when :transfo_done
		@deep_selector.reset_drawing
		@notify_proc.call(event, *args) if @notify_proc
		
	else
		@notify_proc.call(event, *args) if @notify_proc
	end
	
end

#INIT: Initialize the colors
def init_colors
	@color_pal_label = 'khaki'
	@color_pal_value = 'lightgreen'
	@color_pal_hi = 'lightgreen'
	
	@bk_color_pivot_ref = 'orange'
	@fr_color_pivot_ref = 'darkorange'
	@bk_color_pivot_abs = 'yellow'
	@fr_color_pivot_abs = 'gold'	
	
	@color_hovering = 'magenta'
	@color_hovering_trace = 'red'
	@color_trace = 'orange'
end

#INIT: Initialize the text
def init_texts
	@msg_group = T6[:T_TXT_Group]
	@msg_comp = T6[:T_TXT_Component]

	@tip_button_cancel_selection = T6[:TIP_DeepMover_CancelSelection]
	@tip_button_redo_selection = T6[:TIP_DeepMover_RedoSelection]
	@msg_selection_tool = T6[:MSG_DeepMover_Selection]
	@msg_pivot_ref = T6[:MSG_DeepMoverPivotRef]
	@msg_pivot_abs = T6[:MSG_DeepMoverPivotAbs]	
	
	#Tips for screw mode
	@tip_switch_to_pitch = T6[:TIP_SwitchToPitch]
	@tip_switch_to_tpi = T6[:TIP_SwitchToTPI]
	
	#Tips for movement buttons
	@hsh_movement_tips = {}
	@hsh_movement_tips[:movement_selection] = "SELECTION mode: to select one or several objects"
	@hsh_movement_tips[:movement_translation] = "TRANSLATION: defined by direction and offset"
	@hsh_movement_tips[:movement_screw] = "SCREW: combination of Translation and Spin"
	@hsh_movement_tips[:movement_rotation] = "ROTATION: by a pivot of all selected objects"
	@hsh_movement_tips[:movement_spin] = "SPIN: individual rotation of each selected object"
	@hsh_movement_tips[:movement_scale] = "SCALING: by Pivot and Scale Factor"
	@hsh_movement_tips[:movement_explosion] = "EXPLOSION: from Center by a factor"
	
	#Axes Directions
	@tip_vector_z = T6[:T_TIP_Vector_Z] + " [#{T6[:T_KEY_ArrowUp]}]"
	@tip_vector_x = T6[:T_TIP_Vector_X] + " [#{T6[:T_KEY_ArrowRight]}]"
	@tip_vector_y = T6[:T_TIP_Vector_Y] + " [#{T6[:T_KEY_ArrowLeft]}]"
	@tip_vector_no = T6[:T_TIP_Vector_None] + " [#{T6[:T_KEY_ArrowDown]}]"
	@tip_vector_local = T6[:T_TIP_VectorLocal] + " [#{T6[:T_KEY_Tab]}]"	
	@tip_center = T6[:T_TIP_CenterPivot] + " [#{T6[:T_KEY_Tab]}]"	
	
	#Text for interactive change
	@hsh_tip_interactive = {}
	@hsh_tip_interactive[:movement_translation] = "Modify Direction interactively"
	@hsh_tip_interactive[:movement_screw] = @hsh_tip_interactive[:movement_translation]
	@hsh_tip_interactive[:movement_rotation] = "Modify Pivot / Axe interactively"
	@hsh_tip_interactive[:movement_spin] = @hsh_tip_interactive[:movement_rotation]
	
	@hsh_text_interactive = {}
	@hsh_text_interactive[:movement_translation] = "Modify Direction"
	@hsh_text_interactive[:movement_screw] = @hsh_text_interactive[:movement_translation]
	@hsh_text_interactive[:movement_rotation] = "Modify Pivot / Axe"
	@hsh_text_interactive[:movement_spin] = @hsh_text_interactive[:movement_rotation]	
end

#INIT: Return the text for interactive change
def text_interactive ; @hsh_text_interactive[@cur_tool] ; end
def tip_interactive ; @hsh_tip_interactive[@cur_tool] ; end
def text_interactive_length
	len = 0
	@hsh_text_interactive.values.each do |text|
		w, = G6.simple_text_size text
		len = w if len < w
	end
	len + 10
end
	
#INIT: Reset the sequence editor parameters
def reset(keep_sel=false)
	puts "DEEP MOVERE RESET"
	@no_redo = true

	#Reset parameters for movements and selection
	unless keep_sel
		@hsh_selected_comps = {}
		configuration_tool
	end	
	
	#Reset Deep selector
	@deep_selector.reset
	
	#Reset OT Picker
	reset_otpicker
end

#INIT: Reset the OT Picker environment
def reset_otpicker
	@mode = @otpicker.set_mode(:origin)
	@dragging = false
	@pt_origin = @pt_target = nil
	@pivot_info = nil
	@otpicker.reset	
	@otpicker.inference_reset
end

#INIT: Reset the drawing environment
def reset_drawing
	@deep_selector.reset_drawing
end

#INIT: Activate the tool
def activate
	reset
	@invoked = true
end

#INIT: Deactivate the tool
def deactivate
	reset
	vcb_clean
	@invoked = false
end

#-----------------------------------------------
# STATE: Management of interactive state
#-----------------------------------------------

#STATE: Initialize the state for interactive edition
def state_reset
	@dragging = false
	@session_started = false
	@session_stopped = false
	rotation_reset
end

def interactive_mode?
	@session_started && !@session_stopped
end
	
#-----------------------------------------------
# PARAM: Top level methods
#-----------------------------------------------

#PARAM: Set the current tool
def set_tool(tool)
	#Saving the properties of the current tool
	@hsh_params_save[@cur_tool] = @hsh_params
	
	#Registering change of tool
	puts "set tool #{tool}"
	@cur_tool = tool
	
	#Configuring the tool
	configuration_previous
	configuration_tool
	
	#Resetting the OT picker
	@otpicker.reset
	@otpicker.inference_reset
	@otpicker.preselection_set nil
	
	#Filtering the comps to have at least some children for Explosion
	@deep_selector.filter_comps if @cur_tool == :movement_explosion
	
	#Special treatment
	transfer_preselection

	#Notifying the caller
	notify_caller :set_tool, @cur_tool unless @edit_mode
end

#PARAM: Configure the palette environment based on the active tool
def configuration_tool
	state_reset
	@hsh_allowed = {}
	@hsh_allowed[:use_otpicker] = @hsh_allowed[:protractor] = false
	
	case @cur_tool
	
	when :movement_translation, :movement_screw
		@hsh_allowed = { :offset => true, :direction => true, :local => true }
		if @cur_tool == :movement_screw
			@hsh_allowed[:tpi] = @param_use_screw_tpi
			@hsh_allowed[:pitch] = !@hsh_allowed[:tpi]
		end	
		@hsh_allowed[:use_otpicker] = true
		@hsh_params[:direction] = nil
		@hsh_params[:offset] = nil
		#@hsh_params[:pitch] = 1000.mm unless @hsh_params[:pitch]
		#puts "LOCAL = #{@hsh_params.has_key?(:local).inspect}"
		#@hsh_params[:local] = (@cur_tool == :movement_screw)
		@otpicker.remote_enable true
		@otpicker.set_hide_distance false
		
	when :movement_explosion
		@hsh_allowed = { :explode_factor => true }
		@hsh_params[:explode_factor] = 1.0
		@session_stopped = true
		#@dragging = true
		
	when :movement_rotation, :movement_spin
		@hsh_allowed = { :angle => true, :normal => true }
		if @cur_tool == :movement_spin
			@hsh_allowed[:center] = true
			@hsh_allowed[:local] = false
			#@hsh_params[:center] = true
		else
			@hsh_allowed[:pivot_info] = true
		end	
		@hsh_allowed[:use_otpicker] = true
		@hsh_allowed[:protractor] = true
		@hsh_params[:normal] = nil
		@hsh_params[:angle] = 0
		@otpicker.remote_enable false
		@otpicker.set_hide_distance true
	end
end

#PARAM: Configure the current active tool with previous parameters
def configuration_previous
	unless @hsh_params_save[@cur_tool]
		hsh = @hsh_params_save[@cur_tool] = {}
		case @cur_tool
		when :movement_translation
			hsh[:local] = false
		when :movement_screw
			hsh[:local] = true
			hsh[:pitch] = 1.m
		when :movement_spin
			hsh[:center] = true
		end
			
	end	
	@hsh_params.update @hsh_params_save[@cur_tool]	
end

#PARAM: Get and Set methods for translation offset
def param_get_offset ; @hsh_params[:offset] ; end
def param_set_offset(offset)
	@otpicker.reset
	@hsh_params[:offset] = offset
	transformation_from_param_changed
end

#PARAM: Get and Set methods for Angle
def param_get_angle ; @hsh_params[:angle] ; end
def param_set_angle(angle)
	@otpicker.reset
	@hsh_params[:angle] = angle
	transformation_from_param_changed
end

#PARAM: Get and Set methods for screw pitch
def param_get_pitch ; @hsh_params[:pitch] ; end
def param_set_pitch(pitch)
	@hsh_params[:pitch] = pitch
	unless @dragging
		@otpicker.reset
		transformation_from_param_changed
	end	
end
def param_toggle_pitch_mode
	set_screw_tpi !@param_use_screw_tpi
end

#PARAM: Get and Set methods for screw tpi
def param_get_tpi ; 1.0 / @hsh_params[:pitch] ; end
def param_set_tpi(tpi)
	@otpicker.reset
	@hsh_params[:pitch] = 1.0 / tpi
	transformation_from_param_changed
end

def param_ask_screw_parameter
	puts "ask pitch"
end

#PARAM: Get and Set methods for explosion factor
def param_get_explode_factor ; @hsh_params[:explode_factor] ; end
def param_set_explode_factor(explosion_factor)
	@hsh_params[:explode_factor] = explosion_factor
	@dragging = (explosion_factor != 1.0)
	transformation_from_param_changed
end
	
#PARAM: Get and Set methods for Local translation
def param_get_local ; @hsh_params[:local] ; end
def param_toggle_local ; param_set_local !@hsh_params[:local] ; end
def param_set_local(val)
	@hsh_params[:local] = val
	manage_local_axis
	transformation_from_param_changed
end

#PARAM: Get and Set methods for Local translation
def param_get_center ; @hsh_params[:center] ; end
def param_toggle_center ; param_set_center !@hsh_params[:center] ; end
def param_set_center(val)
	@hsh_params[:center] = val
	puts "SET CENTER #{@rot_state}"
	return if @rot_state == :direction
	if val
		@rot_pivot = spin_compute_pivot
		@protractor.set_origin @rot_pivot
	else
		reset
		restore_previous_selection
		@hsh_params[:center] = false
	end	
	manage_local_axis
end

#PARAM: Get and Set methods for translation direction
def param_get_vector_code(type) ; @otpicker.direction_vector_code? ; end
def param_get_vector(type) ; @hsh_params[type] ; end

def param_set_vector_code(code, type)
	vec = direction_vector(code)
	param_set_vector(vec, type) if vec
	@otpicker.option_vector_set_dir(code)
end

def param_set_vector(vec, type)
	@hsh_params[type] = vec
	transformation_from_param_changed
end
	
def param_set_vector_from_arrow
	type = (@hsh_allowed[:normal]) ? :normal : :direction
	code = @otpicker.direction_vector_code?
	param_set_vector_code code, type
end
	
#-----------------------------------------------
# TOP: Top level methods
#-----------------------------------------------

#KEY: Manage key events
def key_manage(event, key, view)
	case event
	
	#SHIFT key
	when :shift_down
		@otpicker.direction_changed
	when :shift_toggle
		if @mode == :origin
			
		else	
			@otpicker.direction_toggle_lock
		end	
	when :shift_up
		@otpicker.direction_changed
	
	#CTRL key	
	when :ctrl_down	
		@otpicker.reset
		@otpicker.inference_reset
		handle_move_zero
		
	when :ctrl_up

	when :ctrl_other_up
	
	#ALT key
	when :alt_down
		@otpicker.no_inference_toggle
	when :alt_up	
		@otpicker.no_inference_set false
		
		
	#Arrows	
	when :arrow_down	
		@otpicker.option_from_arrow key, @keyman.shift_down?, @keyman.ctrl_down?
		param_set_vector_from_arrow
		handle_move_zero
	
	#Backspace	
	when :backspace
		@otpicker.inference_reset

	#TAB	
	when :tab
		if @hsh_allowed[:local]
			param_toggle_local
		elsif @hsh_allowed[:center]	
			param_toggle_center
		end	
		handle_move_zero
		
	when :shift_tab
		
	#Otherkeys
	else
		return false
	
	end
	true
end

#TOP: Notification from OT picker
def notify_from_otpicker(event)
	case event
	when :shift_down
		(@keyman) ? @keyman.shift_down? : false
	when :ctrl_down	
		(@keyman) ? @keyman.ctrl_down? : false
	else
		:no_event
	end	
end

#TOP: Check if in Selection mode - If CTRL is down, then we switch to selection mode
def selection_mode?
	(@cur_tool == :movement_selection || (@deep_selector.selection_empty? || @keyman.ctrl_down?) && @mode != :target)
end

#TOP: Check if in Selection is empty
def selection_empty?
	@deep_selector.selection_empty?
end
	
#TOP: Check if in dragging mode
def dragging? ; @dragging ; end
	
#UNDO: Manage Cancel and Undo
def onCancelUndo(code=nil)
	return true if @deep_selector.cancel_selection	
	false
end
	
#-----------------------------------------------
# SELECTION: Manage selection
#-----------------------------------------------

#SELECTION: Notification from deep selector	
def notify_from_selector(event, *args)
	case event
	when :exclude_comp
		comp, = args
		if @cur_tool == :movement_explosion 
			entities = G6.grouponent_entities comp
			lsubcomps = entities.grep(Sketchup::Group) + entities.grep(Sketchup::ComponentInstance)
			return true if lsubcomps.empty?
		end
		return false
	when :selection_changed, :selection_restore, :selection_uploaded
		@hsh_selected_comps = @deep_selector.get_selected_objects.clone
	when :click_in_void
		puts "notify click in void"
		if selection_mode?
			cancel_selection
			return
		end	
	end	
	
	#Local transformation
	manage_local_axis
	
	notify_caller event, @hsh_selected_comps
	transfer_preselection
end

#SELECTION: Calculate the local axis
def manage_local_axis
	@comp_local = @hsh_selected_comps.values[0]
	if @hsh_selected_comps.empty? || !@hsh_params[:local]
		tr = nil
	elsif @hsh_params[:local]
		comp, tr = @comp_local
		tr = tr * comp.transformation if comp
	end
	@otpicker.option_direction_local_tr tr
	
	local_tr = (@comp_local) ? @comp_local[1] : nil
	@otpicker.set_comp_local_tr local_tr
end

#SELECTION: Transfer the selection to the underlying tool if applicable
def transfer_preselection
	if @hsh_allowed[:use_otpicker]
		puts "transfer selection"
		@otpicker.exclusion_register @hsh_selected_comps.values.collect { |a| a.first }
	end
end
	
#SELECTION: Cancel the selection
def cancel_selection
	return if @deep_selector.cancel_selection
	@otpicker.reset
	@otpicker.inference_reset
	@deep_selector.reset
	configuration_tool
end

#SELECTION: Restore the previous selection
def restore_previous_selection(no_notify=false)
	@deep_selector.restore_previous_selection no_notify
end

#-----------------------------------------------
# MOVE: Mouse move methods
#-----------------------------------------------

#MOVE: top level move method
def handle_move_zero ; handle_move(0, @xmove, @ymove, @view) if @xmove ; end
def handle_move(flags, x, y, view)
	@xmove = x
	@ymove = y
	
	#Selection
	if selection_mode?
		@deep_selector.handle_move(flags, x, y, view)
		return
	end
	#Movement tools
	case @cur_tool
	when :movement_translation, :movement_screw
		translation_mouse_move(flags, x, y, view)
	when :movement_spin, :movement_rotation
		rotation_mouse_move(flags, x, y, view)
	end	
end

#-----------------------------------------------
# CLICK: Mouse click methods
#-----------------------------------------------

#CLICK: Handle Mouse DOWN
def handle_button_down(flags, x, y, view)	
	@time_down_start = Time.now
	@xdown = x
	@ydown = y
	
	#Selection mode
	if selection_mode?
		@deep_selector.handle_button_down(flags, x, y, view)
		return
	end

	#Specific to each tool
	case @cur_tool
	when :movement_translation, :movement_screw
		translation_button_down(flags, x, y, view)

	#Handling rotation
	when :movement_rotation, :movement_spin
		rotation_button_down(flags, x, y, view)	
	end	
end

#CLICK: Handle Mouse UP
def handle_button_up(flags, x, y, view)
	#puts "\nbutton up #{@cur_tool} sel mode = #{@selection_mode}"
	puts "UP mode = #{@mode} dragging = #{@dragging}"
	
	#Interactive session is over
	return if @session_stopped
	
	#Selection mode
	if selection_mode?
		@deep_selector.handle_button_up(flags, x, y, view)
		return
	end
	
	#Specific to each tool
	case @cur_tool
	when :movement_translation, :movement_screw
		translation_button_up(flags, x, y, view)
	when :movement_rotation, :movement_spin
		rotation_button_up(flags, x, y, view)
	end	
end

#CLICK: Handle double click
def handle_doubleclick(flags, x, y, view)
	puts "DEEP MOVER DOUBLE CLICK sel = #{selection_mode?} dragging = #{@dragging}"
	#In selection mode, select the lowest level comp
	if selection_mode? || !@dragging
		@deep_selector.handle_doubleclick(flags, x, y, view)
		return true
		
	#In action mode, repeat last recorded movement if possible	
	elsif @dragging
		puts "DOUBLE CLCIK MOVER dragging = #{@dragging}"
		unless transformation_repeat
			UI.beep
			notify_caller :cancel
		end	
		
	else	
		return false
	end	
	true
end

#CLICK: check if close to initial down position
def surplace?(x, y)
	(Time.now - @time_down_start < 0.8 && (x - @xdown).abs < 3 && (y - @ydown).abs < 3)
end

#---------------------------------------------------------------------------------------------
# TRANSLATION: Interactive control of Translation and Screw
#---------------------------------------------------------------------------------------------

#TRANSLATION: Mouse movement
def translation_mouse_move(flags, x, y, view)
	#puts "handle translation move sel mode = #{@mode}"	

	#Tracking the origin 
	if @mode == :origin
		@origin_info = @otpicker.origin_pick view, x, y, false, false, true
		@pt_origin, @type_origin = @origin_info
		
	#Tracking the target	
	elsif @mode == :target
		return unless @dragging
		@pt_target, = @otpicker.target_pick(view, x, y, true)
		origin = @hsh_params[:origin]
		@hsh_params[:direction] = origin.vector_to @pt_target
		@hsh_params[:offset] = origin.distance @pt_target
		transformation_perform
	end	
end

#TRANSLATION: Handle Mouse DOWN
def translation_button_down(flags, x, y, view)	
	puts "translation down mode = #{@mode} dragging = #{@dragging}"
	if @mode == :origin
		translation_mouse_move(0, x, y, view) unless @pt_origin
		@dragging = true
		@hsh_params[:origin] = @pt_origin
		@mode = @otpicker.set_mode(:target)
	elsif @mode == :target
		#transformation_finish
		@dragging = false
	end	
	@time_down = Time.now
end

#TRANSLATION: Handle Mouse UP
def translation_button_up(flags, x, y, view)
	puts "translation UP mode = #{@mode} dragging = #{@dragging} surplace = #{surplace?(x, y)}"
	if @mode == :target && (!@dragging || !surplace?(x, y))
		transformation_finish
	end	
end

#---------------------------------------------------------------------------------------------
# ROTATION: Interactive control of Rotation and Spin
#---------------------------------------------------------------------------------------------

#ROTATION: Initialize the state for interactive edition
def rotation_reset
	puts "\n*************Rotation reset"
	@rot_state = :direction
	@protractor.reset
	@rot_normal = nil
	@rot_curdir = nil
	@rot_basedir = nil
end

#ROTATION: Handle mouse movement
def rotation_mouse_move(flags, x, y, view)
	@protractor_enabled = true
	@otpicker_no_display = false
	
	#Tracking the origin 
	if @mode == :origin
		@origin_info = @otpicker.origin_pick view, x, y, false, false, true
		@pt_origin, @type_origin = @origin_info
		@ph.do_pick x, y
		
	#Tracking the target	
	elsif @mode == :target
		@pt_target, = @otpicker.target_pick(view, x, y, true)
	end
	
	case @rot_state
	when :direction
		#Calculating the normal and pivot
		if @mode == :origin
			@rot_normal = rotation_compute_normal
			if @hsh_allowed[:pivot_info]
				@pivot_info = rotation_compute_pivot_info		#Rotation
				@rot_pivot, = @pivot_info
			else	
				@rot_pivot = spin_compute_pivot					#Spin
			end	
			
			unless @rot_pivot && @rot_normal
				@protractor_enabled = false
				return
			end	
		
		#No change of mouse
		elsif surplace?(x, y)
			return
			
		#Normal fixed in dragging mode	
		elsif @mode == :target
			@rot_normal = @rot_pivot.vector_to @pt_target
			return unless @rot_normal.valid?
		end
		
		#Storing the information and updating the protractor
		@rot_basedir = @rot_normal.axes[0]
		@protractor.set_placement @rot_pivot, @rot_normal, @rot_basedir

	when :basedir
		#Calculating the base orientation, using inference or not
		pt, @rot_imposed = rotation_point_from_screen(@rot_pivot, @rot_normal)
		vec = @rot_pivot.vector_to pt
		@otpicker_no_display = true unless @rot_imposed
		return unless vec.valid?
		@rot_basedir = vec
		@protractor.set_basedir @rot_basedir
			
	when :angle
		#Calculating the target point, using inference or not
		type_ori = @origin_info[1]
		pt, @rot_imposed = rotation_point_from_screen(@rot_pivot, @rot_normal)
		if @rot_imposed
			@rot_curdir = @rot_pivot.vector_to pt
		else	
			vecdir = @rot_pivot.vector_to pt
			@protractor.check_inference(vecdir, nil, pt)
			@rot_curdir = @protractor.curdir
		end	
		@otpicker_no_display = true unless @rot_imposed
		return unless @rot_curdir.valid?
		
		#Calculating the true angle
		@rot_curdir_prev = @rot_basedir unless @rot_curdir_prev
		dangle = @rot_curdir_prev.angle_between @rot_curdir
		dangle = -dangle if (@rot_curdir_prev * @rot_curdir) % @rot_normal < 0
		@rot_curdir_prev = @rot_curdir
		@hsh_params[:angle] += dangle.radians
		
		#Performing the rotation
		transformation_perform
	end			
end

#ROTATION: Compute the effective pivot (for spin)
def spin_compute_pivot
	comp, tr = @comp_local
	return @pt_origin unless @hsh_params[:center]
	tr * comp.transformation.inverse * comp.bounds.center
end

#ROTATION: Reference component under mouse
def rotation_compute_pivot_info
	pt, type, info_comp = @origin_info
	elt, comp, tr = info_comp
	return [pt] unless comp
	chain_comp = @otpicker.chain_comp.reverse
	return [pt] if chain_comp.empty?
	lcomps = chain_comp.collect { |a| a[0] } 
	index = lcomps.rindex(comp)
	index = 0 unless index
	[pt, chain_comp, index]
end

#ROTATION: Handle Mouse DOWN
def rotation_button_down(flags, x, y, view)	
	case @rot_state
	when :direction
		if @mode == :origin
			@dragging = true
			@mode = @otpicker.set_mode(:target)
			
		elsif @mode == :target
			rotation_direction_stop
		end	
		
	when :basedir	
		@hsh_params[:pivot] = @rot_pivot
		@hsh_params[:normal] = @rot_normal
		@hsh_params[:pivot_info] = @pivot_info
		@rot_state = :angle
		@rot_curdir_prev = nil
		@mode = @otpicker.set_mode(:origin)
		@dragging = true
	
	when :angle
		handle_move(0, x, y, view) unless @dragging || @pt_origin
		@dragging = !@dragging		
	
	end
	
	@time_down = Time.now
end

#ROTATION: Handle Mouse UP
def rotation_button_up(flags, x, y, view)
	case @rot_state
	when :direction
		if Time.now - @time_down_start < 0.8 && (x - @xdown).abs < 3 && (y - @ydown).abs < 3
			puts "continue direction dragging"
		elsif @mode == :target
			rotation_direction_stop
		end	
		@rot_state = :basedir
		@mode = @otpicker.set_mode(:origin)
		
	when :basedir
		if Time.now - @time_down_start < 0.8 && (x - @xdown).abs < 3 && (y - @ydown).abs < 3
			@rot_state = :angle
			@mode = @otpicker.set_mode(:origin)
		end	

	when :angle
		puts "UP angle #{surplace?(x, y)}"
		transformation_finish unless surplace?(x, y) && @dragging
		
	end
end

#ROTATION: Stop input of normal for Rotation and Spin
def rotation_direction_stop
	@dragging = false
	@mode = @otpicker.set_mode(:origin)
	@otpicker.reset
	@otpicker.inference_reset
	
end

#ROTATION: Compute the normal for rotation
def rotation_compute_normal
	#Predefined normal
	direction = @otpicker.direction_forced_any?
	return direction if direction
	
	#Freeze normal when Shift pressed
	return @rot_normal if @keyman.shift_down?
	
	#Normal from the picked face or edge
	face = @ph.picked_face
	edge = @ph.picked_edge
	t = @ph.transformation_at(0)
	if face
		direction = G6.transform_vector(face.normal, t)
	elsif edge
		vec = edge.start.position.vector_to edge.end.position
		direction = G6.transform_vector(vec, t)
	else
		direction = nil
	end
	
	#Normal from the OT Picker
	direction, = @otpicker.direction_pivot_origin unless direction

	direction = Z_AXIS unless direction
	
	#puts "rot direction = #{direction}"
	direction	
end

#ROTATION: Compute the base or angle direction based on the point picked on screen
def rotation_point_from_screen(pivot, direction)
	pt = @pt_origin
	type_ori = @origin_info[1]
	plane = [pivot, direction]
	return [pt, true] if direction && pt.on_plane?(plane) && @otpicker.dof_of(type_ori) < 2
	eye = @view.camera.eye
	line = [eye, eye.vector_to(pt)]
	[Geom.intersect_line_plane(line, plane), false]
end

#--------------------------------------------------------
# TRANSFORMATION : Manage Transformations (generic)
#--------------------------------------------------------

#TRANSFORMATION: Store the initial transformation for selected comps
def store_initial_transformations
	puts "\n*****************STORE TRANSFO\n"
	if @hsh_initial_position_upload
		@hsh_initial_positions = @hsh_initial_position_upload
	else	
		@hsh_initial_positions = @pcompman.save_positions
	end	
	@session_started = true
end

#TRANSFORMATION: Finish a transformation
def transformation_finish
	puts "finish transformation"
	@dragging = false
	@mode = @otpicker.set_mode(:origin)
	@otpicker.reset
	@otpicker.inference_reset

	case @cur_tool
	when :movement_translation, :movement_screw
	
	when :movement_rotation, :movement_spin
		rotation_direction_stop
		rotation_reset
		
	else
		return
	end	
	advise_nominal_transformation
end

#TRANSFORMATION: Check if a movement is already complete for its parameters
def transformation_completed?
	status = true
	case @cur_tool
	when :movement_translation, :movement_screw
		status = false unless @hsh_params[:direction]
	when :movement_rotation, :movement_spin
		#puts "completed = #{@hsh_params.inspect}"
		status = false unless (@hsh_params[:pivot] || @hsh_params[:center]) && @hsh_params[:normal]
	when :movement_explosion
		status = false if @hsh_params[explode_factor] == 1
	end
	status	
end

#TRANSFORMATION: Repeat a transformation
def transformation_repeat
	puts "Repeat transformation"
	tr_info = @hsh_repeat[@cur_tool]
	return false unless tr_info #&& transformation_completed?

	ls_params = []
	
	case @cur_tool
	when :movement_translation
		ls_params.push :offset, :direction, :local
	when :movement_screw
		ls_params.push :offset, :direction, :local, :pitch	
	when :movement_rotation, :movement_spin
		ls_params.push :angle, :pivot, :normal, :center
	when :movement_explosion
		ls_params.push :explode_factor	
	end	
	
	#Setting the last parameters
	return false if ls_params.empty?
	ls_params.each do |param|
		#puts "Param = #{param} info = #{tr_info[param]}"
		@hsh_params[param] = tr_info[param]
	end
	#puts "HSH PARAM = #{@hsh_params.inspect}"
	
	#Repeating the transformation
	transformation_finish
	transformation_perform
	
	true
end

#TRANSFORMATION: Perform the transformation when a parameter is changed
def transformation_from_param_changed
	if @session_stopped
		puts "Transformation REDO no redo = #{@no_redo}"
		transformation_redo
	else
		return false unless transformation_completed?
		transformation_finish
		transformation_perform
	end	
	true
end

#TRANSFORMATION: Reset the environment
def transformation_reset_interactive
	@session_stopped = false
	@deep_selector.reset
	reset_otpicker
	#reset
	restore_previous_selection	
end

#TRANSFORMATION: Perform the transformation
def transformation_perform
	puts "TRANSFORM PERFORM started = #{@session_started}"
	store_initial_transformations unless @session_started
	
	#Buildings the specs for transformation
	@tr_info_cur = transformation_build_specs
	puts "TR INFO SPECS = #{@tr_info_cur}"

	#No transformation
	return false unless @tr_info_cur

	#Executing the movement
	puts "Initial empty" if !@hsh_initial_positions || @hsh_initial_positions.empty?
	puts "Execute single"
	@pcompman.movement_execute_single(@tr_info_cur, @hsh_initial_positions)
	
	#Reinitializing the deep selector drawing context
	@deep_selector.reset_drawing
	
	#Returning OK for transform
	true
end

#TRANSFORMATION: Build the specs for transformation
def transformation_build_specs(final=false)
	tr_info = nil
	
	case @cur_tool
	
	#Translation and Screw
	when :movement_translation, :movement_screw
		offset = @hsh_params[:offset]
		vector = @hsh_params[:direction]
		puts "Offset = #{offset} vector = #{vector}"
		return nil unless vector && vector.valid? && offset
		local = @hsh_params[:local]
		if local
			comp_local, tr_local = @comp_local
			vector = G6.transform_vector(vector, tr_local.inverse)
		end
		tr_info = { :type => @cur_tool, :direction => vector, :offset => offset, :local => local }
		tr_info[:pitch] = @hsh_params[:pitch] if @cur_tool == :movement_screw
	
	when :movement_spin
		comp_local, tr_local = @comp_local
		puts "\nComp local = #{@comp_local}"
		puts "PIVOT = #{@hsh_params[:pivot]}"
		pivot = (@hsh_params[:center]) ? :center : tr_local.inverse * @hsh_params[:pivot]
		normal = @hsh_params[:normal]
		angle = @hsh_params[:angle]
		return nil unless normal && pivot && angle && angle != 0 && normal.valid?
		normal = G6.transform_vector(normal, tr_local.inverse)
		tr_info = { :type => @cur_tool, :normal => normal, :pivot => pivot, :angle => angle }

	when :movement_rotation
		pivot = @hsh_params[:pivot]
		normal = @hsh_params[:normal]
		angle = @hsh_params[:angle]
		return nil unless normal && pivot && angle && angle != 0 && normal.valid?
		
		pt, chain_pivot, icomp = @hsh_params[:pivot_info]
		#puts "\nInfo = #{pt} #{@pivot_info}"
		if chain_pivot && icomp
			comp_ref = chain_pivot[icomp][0]
			tr = chain_pivot[icomp][1] * comp_ref.transformation.inverse
			lst_comp = chain_pivot[icomp..-1].collect { |a| a.first }
			pivot = [tr.inverse * pt, lst_comp]
			normal = G6.transform_vector(normal, tr.inverse)
		else
			pivot = pt
		end	
		tr_info = { :type => @cur_tool, :normal => normal, :pivot => pivot, :angle => angle }
	
	#Explosion	
	when :movement_explosion
		factor = @hsh_params[:explode_factor]
		if factor
			puts "PAS BON EXPLODE #{factor}" if final && factor <= 1.0
			return nil if final && factor <= 1.0
			tr_info = { :type => @cur_tool, :explode_factor => factor }
		end
	end	
	
	#Registering the comps in the transformation specs
	lst = []
	@hsh_selected_comps.each do |id, a|
		comp, tr, chain_comp = a
		lst.push [comp, chain_comp]
	end
	tr_info[:comps] = lst if tr_info
	#tr_info[:comps] = @hsh_selected_comps.values if tr_info
	
	#Registering for further Repeat
	#@hsh_repeat[@cur_tool] = tr_info.clone if tr_info
	
	#returning the movement information
	tr_info
end

#TRANSFORMATION: Redo the transformation (for VCB or API inputs)
def transformation_redo
	return if @no_redo
	return unless transformation_perform
	advise_nominal_transformation
end

#TRANSFORMATION: Notify the caller of the transformation
def advise_nominal_transformation
	#Preparing the information
	hparam = transformation_build_specs true
	
	#Notifying the caller
	notify_caller :transfo_done, hparam
	
	#Registering for further Repeat
	@hsh_repeat[@cur_tool] = hparam.clone if hparam
		
	#End of interactive session
	@session_stopped = true if hparam
end

#TRANSFORMATION: Apply transformation when selection changed if possible
def transformation_apply_at_selection
	puts "APPLY AT SELECTION"
	return unless transformation_perform
	advise_nominal_transformation
end

#TRANSFORMATION: Calculate the vector code from a vector
def direction_code(vec)
	case vec
	when X_AXIS
		:follow_x
	when Y_AXIS
		:follow_y
	when Z_AXIS
		:follow_z
	else
		:no
	end
end

#TRANSFORMATION: Calculate the vector code from a vector
def direction_vector(code)
	case code
	when :follow_x
		X_AXIS
	when :follow_y
		Y_AXIS
	when :follow_z
		Z_AXIS
	else
		nil
	end
end

#-----------------------------------------------
# UPLOAD: Handle Edition mode with upload
#-----------------------------------------------

#UPLOAD: Upload a movement in the parameter bar	
def upload_movement(mvt, hsh_initial_position=nil)
	@edit_mode = mvt
	
	puts "\nUPLOAD MOVEMENT**************************\n"
	#Store the initial positions at upload
	@hsh_initial_position_upload = hsh_initial_position
	@tr_info_cur = mvt
	
	#Upload the comps into the selection
	@deep_selector.upload_selection mvt[:comps]
	
	#Upload the type
	type = mvt[:type]
	set_tool type
	
	#Upload the parameters
	case type
	when :movement_translation, :movement_screw
		@hsh_params[:offset] = mvt[:offset]
		@hsh_params[:direction] = mvt[:direction]
		@hsh_params[:pitch] = mvt[:pitch]
		@hsh_params[:origin] = ORIGIN
	when :movement_explosion
		@hsh_params[:explode_factor] = mvt[:explode_factor]
	when :movement_spin
		@hsh_params[:normal] = mvt[:normal]
		@hsh_params[:pivot] = mvt[:pivot]
		@hsh_params[:center] = :center if @hsh_params[:pivot] == :center
		@hsh_params[:angle] = mvt[:angle]
	end
	
	#Session stopped
	@session_stopped = true
	puts "NO REDO = #{@no_redo}"
	@no_redo = false
end
	
#-----------------------------------------------
# DRAW: Handle drawing for the sequence editor
#-----------------------------------------------

#DRAW: top level drawing method
def draw(view)	
	#Drawing the current selection under mouse
	@deep_selector.draw(view) if selection_mode?
	
	#Drawing the current selection
	factor = (@hsh_allowed[:use_otpicker]) ? 1.0 : 1.05
	@deep_selector.drawing_selection(view, factor) #unless @dragging
	
	unless selection_mode? || @session_stopped
		#Drawing the protractor
		drawing_pivot_info(view)
		drawing_protractor(view) if @hsh_allowed[:protractor] && @protractor_enabled
		
		#Drawing the OT Picker elements
		if @hsh_allowed[:use_otpicker] && !@otpicker_no_display
			@otpicker.draw_all view
		end
	end	
	
	#Drawing the trace of movement in Edit mode
	#if !selection_mode? && @edit_mode #&& !@session_started #&& @session_stopped
	if !selection_mode? && @tr_info_cur &&!@dragging #&& !@session_started #&& @session_stopped
		drawing_movement(view, @tr_info_cur, :trace)
	end
	
	#Drawing the current selection under mouse	
	@deep_selector.post_draw(view) if selection_mode?	
	
	#Message for pivot
	drawing_pivot_message(view) unless selection_mode? || @session_stopped
end

#DRAW: Drawing of protractor and related lines
def drawing_protractor(view)
	@protractor.draw view, 1, @hsh_params[:angle].degrees
	if @rot_basedir && @rot_basedir.valid?
		color = 'black'
		#G6.draw_infinite_line_2d(view, @hsh_params[:pivot], @rot_basedir, color, 1, '_')
		G6.draw_infinite_line_2d(view, @rot_pivot, @rot_basedir, color, 1, '_')
	end	
	if @rot_curdir && @rot_curdir.valid?
		color = 'red'
		G6.draw_infinite_line_2d(view, @rot_pivot, @rot_curdir, color, 1, '_')
		#G6.draw_infinite_line_2d(view, @hsh_params[:pivot], @rot_curdir, color, 1, '_')
	end	
end

#DRAW: Drawing highlights for a movement
def drawing_movement(view, mvt, specs=nil)
	trace_color = comp_color = nil
	case specs
	when :comp
		comp_color = @color_hovering
	
	when :trace
		trace_color = @color_trace
		
	when :trace_comp	
		comp_color = @color_hovering
		trace_color = @color_hovering_trace
	end
	
	mvt[:comps].each do |comp_info|
		next unless comp_info
		comp, chain_comp = comp_info
		G6.grouponent_draw_box(view, comp, comp_color, nil, 1, nil, 1.1, true) if comp_color
		drawing_movement_trace(view, mvt, comp, chain_comp, trace_color) if trace_color
	end	
end

#DRAW: Drawing the informative trace of a movement
def drawing_movement_trace(view, mvt, comp, chain_comp, color=nil)	
	color = @color_trace unless color
	type = mvt[:type]
	
	case type
	
	#Translation, Screw
	when :movement_translation, :movement_screw
		tr_top0 = tr_top_of_comp(comp, chain_comp)
		vec = mvt[:direction]
		vec = G6.transform_vector(vec, tr_top0.inverse) unless mvt[:local]
		offset = mvt[:offset]
		view.line_stipple = ''
		G6.grouponent_tr_top_list(comp).each do |tr_top|
			center = tr_top * comp.transformation.inverse * comp.bounds.center
			v = G6.transform_vector(vec, tr_top)
			prev_center = center.offset v, -offset
			pts2d = [prev_center, center].collect { |pt| view.screen_coords(pt) }
			
			#Line for translation
			view.line_width = 1
			view.drawing_color = color
			view.line_stipple = '_'
			view.draw2d GL_LINE_STRIP, pts2d
			view.line_width = 2
			view.line_stipple = ''
			view.draw GL_LINE_STRIP, [prev_center, center]
			
			#Origin and target arrow
			ptq = G6.pts_circle pts2d[0].x, pts2d[0].y, 3
			view.draw2d GL_POLYGON, ptq
			ptt = G6.pts_triangle(pts2d[1].x, pts2d[1].y, 4, pts2d[0].vector_to(pts2d[1]))
			view.draw2d GL_POLYGON, ptt
		end	
		
	#Spin	
	when :movement_spin
		tr_top0 = tr_top_of_comp(comp, chain_comp)
		pivot = mvt[:pivot]
		normal = mvt[:normal]
		G6.grouponent_tr_top_list(comp).each do |tr_top|
			pivot_top = (pivot == :center) ? tr_top * comp.transformation.inverse * comp.bounds.center : tr_top * pivot
			normal_top = G6.transform_vector(normal, tr_top)
			drawing_rotation_axe(view, pivot_top, normal_top, color)
		end	
		
	#Rotation	
	when :movement_rotation
		pivot_info = hspecs[:pivot]
		pivot = nil
		if pivot_info.class == Array
			pivot0, chain_pivot = pivot_info
			if chain_pivot	
				#Reference comp for pivot of rotation given in local coordinates
				comp_pivot = chain_pivot.first
				tr_top_pivot = tr_top_of_comp(comp_pivot, chain_pivot)

				#Pivot at Model level
				pivot = tr_top_pivot * pivot0
				normal = G6.transform_vector(normal0, tr_top_pivot)
				
			end	
		else
			pivot = pivot_info
		end	
		
	end
end

#DRAW: Draw an axis of rotation. <pivot> and <normal> are given in top level coordinates
def drawing_rotation_axe(view, pivot, normal, color)
	pt1 = pivot.offset normal, -200
	pt2 = pivot.offset normal, 200
	pts2d = [pt1, pt2].collect { |pt| view.screen_coords(pt) }
	
	#Line for normal
	view.line_width = 1
	view.drawing_color = color
	view.line_stipple = '_'
	view.draw2d GL_LINE_STRIP, pts2d
	view.line_width = 2
	view.line_stipple = ''
	view.draw GL_LINE_STRIP, [pt1, pt2]
	
	#Direction arrow
	ptt = G6.pts_triangle(pts2d[1].x, pts2d[1].y, 4, pts2d[0].vector_to(pts2d[1]))
	view.draw2d GL_POLYGON, ptt

end

#DRAW: Compute the Top transformation for a comp and its chain comp
def tr_top_of_comp(comp, chain_comp)
	return chain_comp if chain_comp.class == Geom::Transformation
	tr = comp.transformation.inverse
	chain_comp.each do |c|
		tr = c.transformation * tr
	end
	tr	
end

#DRAW: Draw the pivot information with a reference comp
def drawing_pivot_info(view)
	return if @deep_selector.selection_empty? || !@pivot_info || @pivot_info.length == 1
	pt, chain_comp, icomp = @pivot_info
	return unless pt
	pt2d = view.screen_coords pt
	pts = G6.pts_square pt2d.x, pt2d.y, 6
	if chain_comp && icomp
		comp = chain_comp[icomp][0]
		fr_color = @bk_color_pivot_ref
		view.drawing_color = fr_color
		view.draw2d GL_POLYGON, pts
		G6.grouponent_draw_box(view, comp, fr_color, nil, 3, nil, nil, true)
	elsif !comp
		fr_color = @bk_color_pivot_abs
		view.drawing_color = fr_color
		view.draw2d GL_POLYGON, pts	
	end
end

#DRAW: Drawing the message for rotation reference pivot
def drawing_pivot_message(view)
	return unless @hsh_allowed[:pivot_info] && @pivot_info && @rot_state == :direction && !@deep_selector.selection_empty?
	return if @pivot_info.length == 1
	pt, chain_comp, icomp = @pivot_info
	return unless pt
	if chain_comp && icomp
		comp = chain_comp[icomp][0]
		return if @hsh_selected_comps[comp.entityID]
		G6.draw_rectangle_text(view, @xmove, @ymove, @msg_pivot_ref, @bk_color_pivot_ref, @fr_color_pivot_ref)
	else
		G6.draw_rectangle_text(view, @xmove, @ymove, @msg_pivot_abs, @bk_color_pivot_abs, @fr_color_pivot_abs)
	end	
end

#-----------------------------------------------
# PALETTE: Palette Contribution for dropdown
#-----------------------------------------------
	
#PALETTE: Declare the drop down palette for Comp Selection
def palette_contribution(prefix)
	@deep_selector.palette_contribution(prefix)
end
	
#PALETTE: Floating palette for Movement
def palette_floating_movement(prefix, lst_message_hinfo=nil, docking=:side)
	@palette = Traductor::Palette.get_palette
	
	#Parameters
	@draw_local = self.method("palette_draw_button")
	@prefix = prefix
	@nb_sepa = 0
	@wid_sepa = 8
	
	@prefix = prefix
	
	#Declare the floating palette
	flpal = "pal_floating_#{@prefix}_movement".intern
	hidden_proc = proc { !@invoked }
	hsh_floating = { :hidden_proc => hidden_proc, :docking => docking }
	@palette.declare_floating flpal, hsh_floating
	hsh_def = { :floating => flpal } 
	hdef_prev = @palette.default_button_specs hsh_def

	widbut = 28
	irow = 0
	hshb = { :width => widbut, :height => widbut, :row => irow }
	
	#Buttons for MOVEMENTS
	lst_mov = [:selection, nil, :translation, :screw, :rotation, :spin, :scale, nil, 
	           :explosion, nil, :path, :hide_show, :skeleton]
	lst_mov.each do |symb|
		if symb
			palette_button_movement(hshb, "movement_#{symb}".intern)
		else
			palette_sepa(hshb)
		end
	end
	palette_sepa(hshb)

	#Button: CANCEL SELECTION
	symb = :cancel_selection
	grayed_proc = proc { @deep_selector.selection_empty? || @dragging }
	hsh = { :draw_proc => @draw_local, :tooltip => @tip_button_cancel_selection, :grayed_proc => grayed_proc }
	@palette.declare_button("pal_#{@prefix}_SELECT_#{symb}".intern, hshb, hsh) { cancel_selection }

	#Button: REDO SELECTION
	symb = :redo_selection
	grayed_proc = proc { !@deep_selector.selection_previous? || @dragging }
	hsh = { :draw_proc => @draw_local, :tooltip => @tip_button_redo_selection, :grayed_proc => grayed_proc }
	@palette.declare_button("pal_#{@prefix}_SELECT_#{symb}".intern, hshb, hsh) { restore_previous_selection }

	#PARAMETERS (second Row)
	@hgt_param = 16
	irow = 1
	hshc = { :height => @hgt_param, :row => irow }
	nbt_move = (lst_mov.find_all { |a| a }).length
	nbt_sepa = (lst_mov.find_all { |a| !a }).length
	@widtot_param = (nbt_move + 2) * widbut + (nbt_sepa + 1) * @wid_sepa
	
	#Selection Area
	hidden_proc = proc { @cur_tool != :movement_selection || palette_message? }
	hsh = { :passive => true, :text => @msg_selection_tool, :hidden_proc => hidden_proc, 
	        :bk_color => 'lightgreen', :width => @widtot_param }
	@palette.declare_button("pal_#{@prefix}_selection".intern, hshc, hsh)	
	
	#Custom Messages
	@lst_hidproc_message = []
	if lst_message_hinfo
		lst_message_hinfo.each do |symb, hsh_msg|
			@lst_hidproc_message.push hsh_msg[:hidden_proc]
			hsh = { :width => @widtot_param }
			palette_message_button(symb, hsh_msg, hshc, hsh)
		end
	end
	
	#Parameters
	palette_param_value :offset
	palette_param_vector(:direction)
	palette_param_sepa :pitch
	palette_param_value :pitch
	palette_param_sepa :tpi
	palette_param_value :tpi
	palette_param_value :angle
	palette_param_vector(:normal)
	palette_param_value :explode_factor
	
	#Resetting the default specs
	@palette.default_button_specs hdef_prev
	
	#Return the floating palette symbol
	flpal
end
	
def palette_param_forbidden?(symb)
	!@hsh_allowed[symb] || palette_message?
end
	
#PALETTE: Create a button for message	
def palette_message_button(symb, *hargs)
	@palette.declare_button("pal_#{@prefix}_msg_#{symb}".intern, *hargs)
end
	
def palette_message?
	@lst_hidproc_message.each do |meth|
		return true if meth && !meth.call
	end
	false
end
	
#PALETTE UTIL: Separator
def palette_sepa(hshb)
	@nb_sepa += 1
	hsh = { :draw_proc => :separator_V, :passive => true, :width => @wid_sepa }
	@palette.declare_button("pal_deepmover_SEPA_#{@nb_sepa}".intern, hshb, hsh)
end

#PALETTE UTIL: Buttons for movement tools
def palette_button_movement(hshc, symb)
	bd_color_proc = proc { (symb == :movement_selection && @cur_tool != :movement_selection && selection_mode?) ? ['yellow', 3] : nil }
	value_proc = proc { @cur_tool == symb }
	hsh = { :draw_proc => @draw_local, :value_proc => value_proc, 
	        :tooltip => @hsh_movement_tips[symb], :border_color => bd_color_proc }
	@palette.declare_button("pal_#{@prefix}_MOVE_#{symb}".intern, hshc, hsh) { set_tool symb }
end
	
#**************************************************************************
# PALETTE PARAM: Parameters for Palette Contribution for Composite
#**************************************************************************

#PALETTE PARAM: Buttons for Single value
def palette_param_value(param)
	#Conditions for showing / hiding
	hidden_proc = proc { palette_param_forbidden?(param) }
	hshc = { :hidden_proc => hidden_proc, :height => @hgt_param }
	action_proc = label_proc = tip_label_proc = ask_proc = nil
	
	#Key parameters
	case param
	when :offset
		tip = vprompt = T6[:T_TXT_OffsetModelUnits]
		vtype = :length
		vmin = vmax = vincr = nil
		get_proc = proc { param_get_offset }
		set_proc = proc { |val| param_set_offset val }
		show_proc = proc { |val| Sketchup.format_length(val) }

	when :angle
		tip = vprompt = T6[:T_TXT_AngleDegrees]
		vtype = :float
		vmin = vmax = nil
		vincr = 15
		get_proc = proc { param_get_angle }
		set_proc = proc { |val| param_set_angle val }
		show_proc = proc { |val| sprintf("%0.1f\°", val) }
		
	when :explode_factor
		tip = vprompt = T6[:T_TIP_ExplosionFactor]
		vtype = :float
		vmin = @explode_factor_min
		vmax = @explode_factor_max
		puts "=========================================explode #{vmin} #{vmax}"
		vincr = 0.5
		get_proc = proc { param_get_explode_factor }
		set_proc = proc { |val| param_set_explode_factor val }
		show_proc = proc { |val| sprintf("%2.1f", val) }
	
	when :pitch
		tip = T6[:T_TIP_ScrewPitch] + "\n" + T6[:VCB_HintPitch]
		vprompt = T6[:T_TXT_ScrewPitch]
		vtype = :length
		vmin = vmax = vincr = nil
		get_proc = proc { param_get_pitch }
		set_proc = proc { |val| param_set_pitch val }
		show_proc = proc { |val| Sketchup.format_length(val) }
		#action_proc = proc { param_ask_screw_parameter }
		label_proc = proc { param_toggle_pitch_mode }
		tip_label_proc = proc { (@param_use_screw_tpi) ? @tip_switch_to_pitch : @tip_switch_to_tpi }
		
	when :tpi
		tip = T6[:T_TIP_ScrewTPI] + "\n" + T6[:VCB_HintTPI]
		vprompt = T6[:T_TXT_ScrewTPI]
		vtype = :float
		vmin = vmax = vincr = nil
		get_proc = proc { param_get_tpi }
		set_proc = proc { |val| param_set_tpi val }
		show_proc = proc { |val| sprintf("%0.3f tpi", val) }
		ask_proc = proc { |val| sprintf("%0.3f", val) }
		#action_proc = proc { param_ask_screw_parameter }
		label_proc = proc { param_toggle_pitch_mode }
		tip_label_proc = proc { (@param_use_screw_tpi) ? @tip_switch_to_pitch : @tip_switch_to_tpi }
	end	

	#Common values
	ask_proc = show_proc unless ask_proc
	symb_master = "pal_#{@prefix}_PARAM_#{param}".intern
	wid_field = 70
	widlab = 28
	tip_label_proc = proc { tip } unless tip_label_proc
	
	#Label button
	hsh = { :passive => (label_proc == nil), :draw_proc => @draw_local, :tip_proc => tip_label_proc,
	        :width => widlab, :bk_color => @color_pal_label, :action_proc => label_proc }
	@palette.declare_button("#{symb_master}_label".intern, hshc, hsh)
	
	#Field button
	hshi = { :vtype => vtype, :vprompt => vprompt, :vmin => vmin, :vmax => vmax, :vincr => vincr, 
			 :vsprintf => ask_proc, :vbutsprintf => show_proc, :get_proc => get_proc, :set_proc => set_proc }
	input = Traductor::InputField.new hshi

	hsh = { :input => input, :tooltip => tip, :action_proc => action_proc,
	        :width => wid_field, :bk_color => @color_pal_value  }
	@palette.declare_button("#{symb_master}_value".intern, hshc, hsh)
	
	#Complementary button
	case param
	when :explode_factor
		palette_param_sepa(param)
		wid = @widtot_param - 2 * @wid_sepa - widlab - wid_field
		param_explode_slider("#{symb_master}_slider".intern, hshc, wid, input)
	end
end

#PALETTE PARAM: Separator
def palette_param_sepa(param)
	hidden_proc = proc { palette_param_forbidden?(param) }
	hshc = { :hidden_proc => hidden_proc, :height => @hgt_param }
	palette_sepa(hshc)
end

def param_explode_slider(but_symb, hshc, wid_slider, input)
	hidden_proc = proc { palette_param_forbidden?(:explode_factor) }
	hshc = { :hidden_proc => hidden_proc, :height => @hgt_param }
	
	#Progression bar
	#hspecs = { :notify_proc => self.method("animation_explode_slider_proc") }
	hspecs = { :input => input }
	hsh = { :width => wid_slider, :tooltip => "tip slider explode" }
	@palette.declare_slider_button(but_symb, hspecs, hshc, hsh)
end

#PALETTE: Control method for the slider of clip animation
def animation_explode_slider_proc(code, *args)
	case code
	when :value_min
		@explode_factor_min
	when :value_max	
		@explode_factor_max
	when :current_value
		param_get_explode_factor
	when :double_click
		puts "double click Explode"
	when :set_value
		factor, = args
		param_set_explode_factor factor
	when :format_value
		factor, = *args
		"#{sprintf("%2.1f", factor)}"
	else
		nil
	end
end

#PALETTE PARAM: Buttons for translation vector
def palette_param_vector(param)
	#Type of parameters
	case param
	when :direction
		main_tip = T6[:TIP_DeepMover_DirectionTranslation]
	when :normal
		main_tip = T6[:TIP_DeepMover_NormalRotation]	
	end
	
	#Conditions for showing / hiding
	hidden_proc = proc { palette_param_forbidden?(param) }
	hshc = { :hidden_proc => hidden_proc, :height => @hgt_param }
	palette_sepa(hshc)
	
	#Common values
	widlab = 24
	symb_master = "pal_composite_PARAM_#{param}".intern
	
	#Label button
	hsh = { :passive => true, :draw_proc => @draw_local, :tooltip => main_tip,
	        :width => widlab, :bk_color => @color_pal_label }
	@palette.declare_button("#{symb_master}_label".intern, hshc, hsh)	

	#Creating the buttons for Axis
	hshp = { :width => @hgt_param, :hi_color => 'white' }
	
	value_proc = proc { param_get_vector_code(param) == :no }
	hsh = { :text => 'N', :tooltip => @tip_vector_no, :bk_color => 'lavender', :value_proc => value_proc  }
	@palette.declare_button("#{symb_master}__no".intern, hshc, hshp, hsh) { param_set_vector_code(:no, param) }

	palette_sepa hshc
	value_proc = proc { param_get_vector_code(param) == :follow_x }
	hsh = { :text => 'X', :tooltip => @tip_vector_x, :bk_color => 'tomato', :value_proc => value_proc  }
	@palette.declare_button("#{symb_master}__x".intern, hshc, hshp, hsh) { param_set_vector_code(:follow_x, param) }

	value_proc = proc { param_get_vector_code(param) == :follow_y }
	hsh = { :text => 'Y', :tooltip => @tip_vector_y, :bk_color => 'limegreen', :value_proc => value_proc  }
	@palette.declare_button("#{symb_master}__y".intern, hshc, hshp, hsh) { param_set_vector_code(:follow_y, param) }

	value_proc = proc { param_get_vector_code(param) == :follow_z }
	hsh = { :text => 'Z', :tooltip => @tip_vector_z, :bk_color => 'skyblue', :value_proc => value_proc  }
	@palette.declare_button("#{symb_master}__z".intern, hshc, hshp, hsh) { param_set_vector_code(:follow_z, param) }	
	
	#Creating the buttons for Local
	wid = @hgt_param
	#wid = 24
	hshp = { :width => wid, :hi_color => @color_pal_hi }
	
	palette_sepa hshc
	hidden_proc2 = proc { hidden_proc.call || !@hsh_allowed[:local] }
	value_proc = proc { param_get_local }
	hsh = { :draw_proc => @draw_local, :tooltip => @tip_vector_local,
	        :value_proc => value_proc, :hidden_proc => hidden_proc2 }
	@palette.declare_button("#{symb_master}__local".intern, hshc, hshp, hsh) { param_toggle_local }	
	puts "PAL LOCAL = #{"#{symb_master}__local".intern}"
	
	#Creating the buttons for Center
	palette_sepa hshc
	hidden_proc2 = proc { hidden_proc.call || !@hsh_allowed[:center] }
	value_proc = proc { param_get_center }
	hsh = { :draw_proc => @draw_local, :tooltip => @tip_center, 
	        :value_proc => value_proc, :hidden_proc => hidden_proc2 }
	@palette.declare_button("#{symb_master}__center".intern, hshc, hshp, hsh) { param_toggle_center }		
end

#**************************************************************************
# PALETTE OPENGL: Icon methods
#**************************************************************************

#PALETTE OPENGL: Drawing methods for composite palette
def palette_draw_button(symb, dx, dy, main_color, frame_color, selected, grayed)
	lst_gl = []
	dx2 = dx / 2
	dy2 = dy / 2

	case symb.to_s
			
	#Process Management buttons
	when /SELECT_(.+)/
		symb_action = $1.intern
		
		bkcol = 'lavender'
		frcol = 'black'
		wid = (selected) ? 2 : 1
		
		pts = G6.pts_round_rectangle 0, 0, dx, dy, 4
		lst_gl.push [GL_POLYGON, pts, bkcol]
		lst_gl.push [GL_LINE_LOOP, pts, frcol, wid, '']
		
		case symb_action
		
		when :cancel_selection, :redo_selection
			color = (grayed) ? 'lightgrey' : 'green'
			frame_color = ((grayed)) ? 'black' : 'yellow'
			dx2 = dx / 2
			dy2 = dy / 2
			y = dy2
			ptmid = Geom::Point3d.new dx2, y
			t = Geom::Transformation.rotation ptmid, Z_AXIS, 30.degrees
			b = 6
			lpti = [[dx2-b, y], [dx2, dy], [dx2+b, y], [dx2, y+2]]
			pts = lpti.collect { |a| t * Geom::Point3d.new(*a) }
			lst_gl.push [GL_POLYGON, pts, color]
			lst_gl.push [GL_LINE_LOOP, pts, frame_color, 1, '']
			
			b = 4
			lpti = [[dx2-2, y+2], [dx2+2, y+2], [dx2+2, y-b], [dx2-2, y-b]]
			pts = lpti.collect { |a| t * Geom::Point3d.new(*a) }
			lst_gl.push [GL_POLYGON, pts, color]
			lst_gl.push [GL_LINE_STRIP, pts[1..-1], frame_color, 1, '']
			
			if symb_action == :cancel_selection
				color = (grayed) ? 'gray' : 'red'
				dec = dx2
				pts = G6.pts_rectangle dx - dec, 0, dec, dec
				lst_gl.push [GL_LINES, [pts[0], pts[2], pts[1], pts[3]], color, 2, '']
			
			else
				color = (grayed) ? 'gray' : 'blue'
				dec = 2
				xc = dx - dx / 4 - 2
				yc = dy / 4
				radius = dx / 4 - 2
				n = 12
				
				pts = []
				for i in -5..10
					angle = i * Math::PI / n 
					pts.push Geom::Point3d.new(1 + xc + radius * Math.sin(angle), yc + radius * Math.cos(angle))
				end	
				xend = pts.first.x
				yend = pts.first.y
				pts2 = []
				pts2.push Geom::Point3d.new(xend - dec, yend - dec)
				pts2.push Geom::Point3d.new(xend - dec, yend + dec)
				pts2.push Geom::Point3d.new(xend + dec, yend - dec)
				
				pts.each { |pt| pt.x = 2 * xc - pt.x + 1 }
				pts2.each { |pt| pt.x = 2 * xc - pt.x + 1 }
				lst_gl.push [GL_LINE_STRIP, pts, color, 2, '']
				lst_gl.push [GL_POLYGON, pts2, color]			
			end
		
		end
		
	#Movement buttons	
	when /MOVE_(.+)/
		symb_move = $1.intern
		return lst_gl if grayed
		type = (selected) ? nil : :seq
		type = :seq
		lst_gl.concat @ogl.picto_opengl(symb_move, type, false, dx, dy, nil, selected)
				
	when /PARAM_(.+)_label/
		symb_param = $1.intern
		bcolor = 'blue'
		
		case symb_param
		when :offset
			lines = [Geom::Point3d.new(0, dy2), Geom::Point3d.new(dx, dy2)]
			lst_gl.push [GL_LINE_STRIP, lines, bcolor, 2, '']
			
			lines = []
			dec = 5
			nx = dx / dec
			ymin = dy2 - 4
			ymax = dy2 + 4
			for i in 0..nx
				x = i * dec
				lines.push Geom::Point3d.new(x, ymin), Geom::Point3d.new(x, ymax)
			end	
			lst_gl.push [GL_LINES, lines, bcolor, 1, '']
				
		when :direction
			t = Geom::Transformation.rotation Geom::Point3d.new(dx2, dy2), Z_AXIS, 15.degrees
			dec = 6
			lines = [Geom::Point3d.new(0, dy2), Geom::Point3d.new(dx-dec, dy2)].collect { |pt| t * pt }
			lst_gl.push [GL_LINE_STRIP, lines, bcolor, 3, '']
			
			pt1 = Geom::Point3d.new(dx, dy2)
			pt2 = Geom::Point3d.new(dx-dec, dy2-dec)
			pt3 = Geom::Point3d.new(dx-dec, dy2+dec)
			pts = [pt1, pt2, pt3].collect { |pt| t * pt }
			lst_gl.push [GL_POLYGON, pts, bcolor]

		when :angle
			dec = 1
			origin = Geom::Point3d.new dec, dec
			vec1 = Geom::Vector3d.new(dx - 2 * dec, 0, 0)
			vec2 = Geom::Vector3d.new(dx2, dy-dec, 0)
			pt1 = origin.offset vec1
			pt2 = origin.offset vec2
			len = dx2 * 1.3
			pt1a = origin.offset vec1, len
			pt2a = origin.offset vec2, len
			pts = G6.pts_pie origin, pt1a, pt2a
			lst_gl.push [GL_POLYGON, pts, 'red']
			lst_gl.push [GL_LINE_STRIP, pts[1..-1], 'gray', 1, '']
			lst_gl.push [GL_LINE_STRIP, [pt1, origin, pt2], bcolor, 2, '']

		when :normal
			dec = 1
			line = [Geom::Point3d.new(dx2, 0), Geom::Point3d.new(dx2, dy)]
			ltri = 4
			arrow = [Geom::Point3d.new(dx2-ltri, dy-ltri), Geom::Point3d.new(dx2+ltri, dy-ltri), line.last]
			lst_gl.push [GL_POLYGON, arrow, bcolor]
			lst_gl.push [GL_LINE_STRIP, line, bcolor, 2, '']
			r = 5
			pts = G6.pts_circle dx2, dy2-1, r, 12
			t = Geom::Transformation.scaling Geom::Point3d.new(dx2, dy2), 1, 0.7, 1
			pts = pts[3..9].collect { |pt| t * pt }
			lst_gl.push [GL_LINE_STRIP, pts, 'green', 2, '']
			pt = pts.first
			pts = G6.pts_triangle pt.x-1, pt.y, 2
			lst_gl.push [GL_POLYGON, pts, 'green']

		when :tpi
			dec = 1
			line = [Geom::Point3d.new(dec, dy2), Geom::Point3d.new(dx-dec, dy2)]
			ltri = 4
			arrow = [Geom::Point3d.new(dx-ltri-dec, dy2-ltri), Geom::Point3d.new(dx-ltri-dec, dy2+ltri), line.last]
			lst_gl.push [GL_POLYGON, arrow, bcolor]
			lst_gl.push [GL_LINE_STRIP, line, bcolor, 3, '']
			r = 5
			t = Geom::Transformation.scaling Geom::Point3d.new(dx2, dy2), 0.7, 1, 1
			for x in [3, dx2, dx-3]
				pts = G6.pts_circle x, dy2, r, 12
				pts = pts[6..-1].collect { |pt| t * pt }
				lst_gl.push [GL_LINE_STRIP, pts, 'green', 2, '']
			end

		when :pitch
			dec = 0
			ltri = 3
			d = 4
			line = [Geom::Point3d.new(0, dy2), Geom::Point3d.new(dx, dy2)]
			lst_gl.push [GL_LINE_STRIP, line, 'lightgrey', dx2, '']
			lines = []
			x1, x2 = [1, dx-2*d]
			for x in [x1, x2]
				lines.push Geom::Point3d.new(x, 0), Geom::Point3d.new(x+d, dy)
			end
			lst_gl.push [GL_LINES, lines, 'green', 4, '']
			lst_gl.push [GL_LINES, lines, 'blue', 2, '']
			line = [Geom::Point3d.new(d+x1, dy2), Geom::Point3d.new(x2, dy2)]
			arrow = [Geom::Point3d.new(x2-d-ltri, dy2-ltri), Geom::Point3d.new(x2-d-ltri, dy2+ltri), line.last]
			lst_gl.push [GL_POLYGON, arrow, bcolor]
			lst_gl.push [GL_LINE_STRIP, line, bcolor, 1, '']
			
		when :explode_factor
			dec = 1
			dp = 2
			lines = []
			arrows = []
			lines.push Geom::Point3d.new(dx2-dec, dy2+dec), Geom::Point3d.new(dec, dy-dec)
			arrows.concat G6.pts_square lines.last.x, lines.last.y, dp
			lines.push Geom::Point3d.new(dx2-dec, dy2-dec), Geom::Point3d.new(dec, dec)
			arrows.concat G6.pts_square lines.last.x, lines.last.y, dp
			lines.push Geom::Point3d.new(dx2+dec, dy2+dec), Geom::Point3d.new(dx-dec, dy-dec)
			arrows.concat G6.pts_square lines.last.x, lines.last.y, dp
			lines.push Geom::Point3d.new(dx2+dec, dy2-dec), Geom::Point3d.new(dx-dec, dec)
			arrows.concat G6.pts_square lines.last.x, lines.last.y, dp
			
			lst_gl.push [GL_QUADS, arrows, 'red']
			lst_gl.push [GL_LINES, lines, bcolor, 2, '']
			
		end
	
	#Option buttons
	when /__local/
		t = Geom::Transformation.rotation Geom::Point3d.new(dx2, 0), Z_AXIS, 15.degrees
		dec = 2
		len = dy - dec - 1
		wid = 1
		if selected
			wid = 3
		end	
		
		center = Geom::Point3d.new(dec, dec)
		ptx = center.offset X_AXIS, len
		pty = center.offset X_AXIS + Y_AXIS, len
		ptz = center.offset Y_AXIS, len
		lst_gl.push [GL_LINE_STRIP, [center, ptx], 'red', wid, '']
		lst_gl.push [GL_LINE_STRIP, [center, pty], 'green', wid, '']
		lst_gl.push [GL_LINE_STRIP, [center, ptz], 'blue', wid, '']

	when /__center/
		r = [dx2, dy2].min
		color = (selected) ? 'red' : 'black'
		dec = 3
		wid = (selected) ? 2 : 1
		pts = G6.pts_circle dx2, dy2, r-1
		lst_gl.push [GL_LINE_LOOP, pts, color, wid, '']
		lpti = [[dx2-dec, dy2], [dx2+dec, dy2], [dx2, dy2-dec], [dx2, dy2+dec]]
		pts = lpti.collect { |x, y| Geom::Point3d.new(x, y) }
		lst_gl.push [GL_LINES, pts, color, wid, '']
	end
	
	lst_gl	
end
	
#--------------------------------------------------
# MENU: Contextual menu
#--------------------------------------------------

#Contextual menu
def menu_contribution(cxmenu)
	cxmenu.add_sepa
	
	#Selection mode
	if selection_mode?
		@deep_selector.menu_contribution cxmenu
		
		unless selection_empty?
			cxmenu.add_sepa
			cxmenu.add_item(T6[:TIP_DeepMover_CancelSelection]) { cancel_selection }
		end	
	end
	
	case @cur_tool
	when :movement_rotation
		menu_pivot_ref(cxmenu)
	end
	
	#Movement repeat
	if !selection_mode? && @hsh_repeat[@cur_tool]
		cxmenu.add_sepa
		text = T6[:TIP_DeepMover_Repeat] + " [#{T6[:T_KEY_DoubleClick]}]"
		cxmenu.add_item(text) { transformation_repeat }
	end	
end

#SELECT: Calculate the drop down list
def menu_pivot_ref(cxmenu)
	return unless @pivot_info
	pt, chain_comp, icomp = @pivot_info
	return unless chain_comp
	
	#Building the drop down list 
	msg = T6[:MNU_DeepMoverPivotRef]
	chain_comp.each_with_index do |comp_tr, j|
		comp, tr = comp_tr
		name = comp.name
		if name.empty?
			name = (comp.instance_of?(Sketchup::Group)) ? @msg_group : @msg_comp
		end	
		label = "#{msg} #{j} - #{name}"
		label = ">> " + label if j == icomp
		cxmenu.add_item(label) {@pivot_info[2] = j }
	end	
	label = "#{msg} #{T6[:T_TXT_Model]}"
	label = ">> " + label unless icomp
	cxmenu.add_item(label) {@pivot_info[2] = nil }
end
	
#-----------------------------------------------
# VCB: VCB Management
#-----------------------------------------------

#VCB: Standard method to accept text in the VCB: Offset distance and nb of segment
def onUserText(text, view)	
	puts "Deep Selector user text = #{text} tool = #{@cur_tool.inspect}"
	vcb = @hsh_vcb[@cur_tool]
	puts "VCB = #{vcb}"
	return UI.beep unless vcb
	action_from_VCB(vcb, @cur_tool) if vcb.process(text, @palette)
end

#VCB: Initialize VCB
def vcb_init
	vcb_message_init
	@hsh_vcb = {}
	
	#Translation
	vcb = @hsh_vcb[:movement_translation] = Traductor::VCB.new
	vcb.declare_input_format :offset, "l"
	
	#Screw
	vcb = @hsh_vcb[:movement_screw] = Traductor::VCB.new
	vcb.declare_input_format :offset, "l"
	vcb.declare_input_format :pitch, "l__*"
	vcb.declare_input_format :tpi, "f__/"
	
	#Rotation
	vcb = @hsh_vcb[:movement_rotation] = @hsh_vcb[:movement_spin] = Traductor::VCB.new
	vcb.declare_input_format :angle_deg, "f"
	vcb.declare_input_format :angle, "a"
	vcb.declare_input_format :rotation180plus, "_++"
	vcb.declare_input_format :rotation180minus, "_--"
	vcb.declare_input_format :rotation90plus, "_+"
	vcb.declare_input_format :rotation90minus, "_-"
	
	#Explosion
	vcb = @hsh_vcb[:movement_explosion] = Traductor::VCB.new
	vcb.declare_input_format :factor, "f"
end

#VCB: Execute actions from the VCB inputs	
def action_from_VCB(vcb, cur_tool)
	puts "\nVCB ACTION STOPPED = #{@session_stopped}"
	
	case cur_tool

	#Translation, Screw
	when :movement_translation, :movement_screw
		offset = pitch = tpi = nil
		vcb.each_result do |symb, val, suffix|
			case symb
			when :offset
				offset = val
			when :pitch
				pitch = val if cur_tool == :movement_screw
			when :tpi
				tpi = val if cur_tool == :movement_screw
			end
		end
		puts "VCB Offset = #{offset} pitch = #{pitch} tpi = #{tpi}"
		param_set_offset offset if offset
		param_set_pitch pitch if pitch
		param_set_pitch tpi if tpi

	#Rotation, Spin
	when :movement_rotation, :movement_spin
		angle = nil
		vcb.each_result do |symb, val, suffix|
			case symb
			when :angle_deg
				angle = val
			when :angle
				angle = val
			when :rotation180plus
				angle = 180
			when :rotation180minus
				angle = -180
			when :rotation90plus
				angle = 90
			when :rotation90minus
				angle = -90
			end
		end
		puts "VCB Angle = #{angle}"
		param_set_angle angle if angle

	#Explosion
	when :movement_explosion
		factor = nil
		vcb.each_result do |symb, val, suffix|
			case symb
			when :factor
				factor = val
			end
		end
		puts "VCB factor = #{factor}"
		if factor && factor >= 1 && factor <= @explode_factor_max
			param_set_explode_factor(factor)
		else
			UI.beep
		end	
		
	end
end	
	
#VCB: Initialize the messages for the status text bar
def vcb_message_init
	@msg_vcb_selection_header = "SELECTION"
	@msg_vcb_selection_select = "Click on objects and select level in drop-down"
	@msg_vcb_selection_double = "Double-Click to select object at lowest level"
	
	@msg_vcb_ctrl_selection = "CTRL to modify Selection"
	@msg_vcb_cx_menu = "Context Menu for more"
	
	@msg_vcb_explosion_header = "EXPLOSION"
	@msg_vcb_explosion_factor = T6[:T_TXT_Factor]
	@msg_vcb_explosion = "Enter the explosion factor in VCB or Palette"

	@msg_vcb_rotation_angle = T6[:T_TXT_Angle]
	@msg_vcb_rotation_header = "ROTATION"
	@msg_vcb_spin_header = "SPIN"

	@msg_vcb_translation_offset = T6[:T_TXT_Offset]
	@msg_vcb_translation_header = "TRANSLATION"
	@msg_vcb_screw_header = "SCREW"

	@msg_vcb_scale_header = "SCALING"
end
	
#VCB: Manage the Sketchup Status bar area
def vcb_message	
	#Selection Mode
	return vcb_message_selection if selection_mode?
	
	#Assembling the text message
	cxmenu = true
	lmsg = []
	case @cur_tool

	when :movement_translation
		lmsg.push @msg_vcb_translation_header
	
	when :movement_screw
		lmsg.push @msg_vcb_screw_header

	when :movement_rotation
		cxmenu = true
		lmsg.push @msg_vcb_rotation_header

	when :movement_spin
		lmsg.push @msg_vcb_spin_header
		
	when :movement_explosion
		lmsg.push @msg_vcb_explosion_header, @msg_vcb_explosion

	when :movement_scale
		lmsg.push @msg_vcb_scale_header
		
	end	

	#Calculating the label and value
	svalue = label = ""	
	case @cur_tool

	when :movement_translation, :movement_screw
		label = @msg_vcb_translation_offset
		offset = @hsh_params[:offset]
		svalue = (offset) ? Sketchup.format_length(offset) : ""

	when :movement_rotation, :movement_spin
		if @rot_state == :angle
			label = @msg_vcb_rotation_angle
			angle = @hsh_params[:angle]
			svalue = (angle) ? sprintf("%0.1f\°", angle) : ""	
		else
			label = ""
		end
		
	when :movement_explosion
		cxmenu = false
		label = @msg_vcb_explosion_factor
		factor = @hsh_params[:explode_factor]
		svalue = (factor) ? sprintf("%2.1f", factor) : ""
	end	
	
	#Showing the messages
	lmsg.push @msg_vcb_cx_menu if cxmenu
	lmsg.push @msg_vcb_ctrl_selection
	msg = lmsg.join " - "
	Sketchup::set_status_text msg
	Sketchup::set_status_text label, SB_VCB_LABEL
	Sketchup::set_status_text svalue, SB_VCB_VALUE	
end
	
#VCB: Clean the VCB and status bar area
def vcb_clean
	Sketchup::set_status_text ''
	Sketchup::set_status_text '', SB_VCB_LABEL
	Sketchup::set_status_text '', SB_VCB_VALUE	
end
	
def vcb_message_selection
	lmsg = []
	lmsg.push @msg_vcb_selection_header, @msg_vcb_selection_select, @msg_vcb_selection_double
	
	msg = lmsg.join " - "
	label = "Selected"
	nb = @deep_selector.get_selected_objects.length
	svalue = (nb > 1) ? "#{nb} Objects" : "#{nb} Object"

	Sketchup::set_status_text msg
	Sketchup::set_status_text label, SB_VCB_LABEL
	Sketchup::set_status_text svalue, SB_VCB_VALUE		
end
	
end	#class DeepMover

end	#End Module Traductor